def main():
    """Entry point for the application script"""
    print("what do you wnt")
    print("Call your main application code here")
    import geigerlog
    print("Do what you want")


if __name__ == "__main__":
    main()

